import json

from stream_chat.base.channel import ChannelInterface, add_user_id


class Channel(ChannelInterface):
    async def send_message(self, message, user_id, **options):
        """
        Send a message to this channel

        :param message: the Message object
        :param user_id: the ID of the user that created the message
        :return: the Server Response
        """
        payload = {"message": add_user_id(message, user_id), **options}
        return await self.client.post(f"{self.url}/message", data=payload)

    async def send_event(self, event, user_id):
        """
        Send an event on this channel

        :param event: event data, ie {type: 'message.read'}
        :param user_id: the ID of the user sending the event
        :return: the Server Response
        """
        payload = {"event": add_user_id(event, user_id)}
        return await self.client.post(f"{self.url}/event", data=payload)

    async def send_reaction(self, message_id, reaction, user_id):
        """
        Send a reaction about a message

        :param message_id: the message id
        :param reaction: the reaction object, ie {type: 'love'}
        :param user_id: the ID of the user that created the reaction
        :return: the Server Response
        """
        payload = {"reaction": add_user_id(reaction, user_id)}
        return await self.client.post(f"messages/{message_id}/reaction", data=payload)

    async def delete_reaction(self, message_id, reaction_type, user_id):
        """
        Delete a reaction by user and type

        :param message_id: the id of the message from which te remove the reaction
        :param reaction_type: the type of reaction that should be removed
        :param user_id: the id of the user
        :return: the Server Response
        """
        return await self.client.delete(
            f"messages/{message_id}/reaction/{reaction_type}",
            params={"user_id": user_id},
        )

    async def create(self, user_id):
        """
        Create the channel

        :param user_id: the ID of the user creating this channel
        :return:
        """
        self.custom_data["created_by"] = {"id": user_id}
        return await self.query(watch=False, state=False, presence=False)

    async def query(self, **options):
        """
        Query the API for this channel, get messages, members or other channel fields

        :param options: the query options, check docs on https://getstream.io/chat/docs/
        :return: Returns a query response
        """
        payload = {"state": True, "data": self.custom_data, **options}

        url = f"channels/{self.channel_type}"
        if self.id is not None:
            url = f"{url}/{self.id}"

        state = await self.client.post(f"{url}/query", data=payload)

        if self.id is None:
            self.id = state["channel"]["id"]

        return state

    async def query_members(self, filter_conditions, sort=None, **options):
        """
        Query the API for this channel to filter, sort and paginate its members efficiently.

        :param filter_conditions: filters, checks docs on https://getstream.io/chat/docs/
        :param sort: sorting field and direction slice, check docs on https://getstream.io/chat/docs/
        :param options: pagination or members based channel searching details
        :return: Returns members response

        eg.
        channel.query_members(filter_conditions={"name": "tommaso"},
                              sort=[{"created_at": -1}, {"updated_at": 1}],
                              offset=0,
                              limit=10)
        """

        payload = {
            "id": self.id,
            "type": self.channel_type,
            "filter_conditions": filter_conditions,
            "sort": self.client.normalize_sort(sort),
            **options,
        }
        response = await self.client.get(
            "members", params={"payload": json.dumps(payload)}
        )
        return response["members"]

    async def update(self, channel_data, update_message=None):
        """
        Edit the channel's custom properties

        :param channel_data: the object to update the custom properties of this channel with
        :param update_message: optional update message
        :return: The server response
        """
        payload = {"data": channel_data, "message": update_message}
        return await self.client.post(self.url, data=payload)

    async def update_partial(self, to_set=None, to_unset=None):
        """
        Update channel partially

        :param to_set: a dictionary of key/value pairs to set or to override
        :param to_unset: a list of keys to clear
        """
        payload = {"set": to_set or {}, "unset": to_unset or []}
        return await self.client.patch(self.url, data=payload)

    async def delete(self):
        """
        Delete the channel. Messages are permanently removed.

        :return: The server response
        """
        return await self.client.delete(self.url)

    async def truncate(self):
        """
        Removes all messages from the channel

        :return: The server response
        """
        return await self.client.post(f"{self.url}/truncate")

    async def add_members(self, user_ids):
        """
        Adds members to the channel

        :param user_ids: user IDs to add as members
        :return:
        """
        return await self.client.post(self.url, data={"add_members": user_ids})

    async def invite_members(self, user_ids):
        """
        invite members to the channel

        :param user_ids: user IDs to invite
        :return:
        """
        return await self.client.post(self.url, data={"invites": user_ids})

    async def add_moderators(self, user_ids):
        """
        Adds moderators to the channel

        :param user_ids: user IDs to add as moderators
        :return:
        """
        return await self.client.post(self.url, data={"add_moderators": user_ids})

    async def remove_members(self, user_ids):
        """
        Remove members from the channel

        :param user_ids: user IDs to remove from the member list
        :return:
        """
        return await self.client.post(self.url, data={"remove_members": user_ids})

    async def demote_moderators(self, user_ids):
        """
        Demotes moderators from the channel

        :param user_ids: user IDs to demote
        :return:
        """
        return await self.client.post(self.url, data={"demote_moderators": user_ids})

    async def mark_read(self, user_id, **data):
        """
        Send the mark read event for this user, only works if the `read_events` setting is enabled

        :param user_id: the user ID for the event
        :param data: additional data, ie {"message_id": last_message_id}
        :return: The server response
        """
        payload = add_user_id(data, user_id)
        return await self.client.post(f"{self.url}/read", data=payload)

    async def get_replies(self, parent_id, **options):
        """
        List the message replies for a parent message

        :param parent_id: The message parent id, ie the top of the thread
        :param options: Pagination params, ie {limit:10, id_lte: 10}
        :return: A response with a list of messages
        """
        return await self.client.get(f"messages/{parent_id}/replies", params=options)

    async def get_reactions(self, message_id, **options):
        """
        List the reactions, supports pagination

        :param message_id: The message id
        :param options: Pagination params, ie {"limit":10, "id_lte": 10}
        :return: A response with a list of reactions
        """
        return await self.client.get(f"messages/{message_id}/reactions", params=options)

    async def ban_user(self, target_id, **options):
        """
        Bans a user from this channel

        :param target_id: the ID of the user to ban
        :param options: additional ban options, ie {"timeout": 3600, "reason": "offensive language is not allowed here"}
        :return: The server response
        """
        return await self.client.ban_user(
            target_id, type=self.channel_type, id=self.id, **options
        )

    async def unban_user(self, target_id, **options):
        """
        Removes the ban for a user on this channel

        :param target_id: the ID of the user to unban
        :return: The server response
        """
        return await self.client.unban_user(
            target_id, type=self.channel_type, id=self.id, **options
        )

    async def accept_invite(self, user_id, **data):
        payload = add_user_id(data, user_id)
        payload["accept_invite"] = True
        response = await self.client.post(self.url, data=payload)
        self.custom_data = response["channel"]
        return response

    async def reject_invite(self, user_id, **data):
        payload = add_user_id(data, user_id)
        payload["reject_invite"] = True
        response = await self.client.post(self.url, data=payload)
        self.custom_data = response["channel"]
        return response

    async def send_file(self, url, name, user, content_type=None):
        return await self.client.send_file(
            f"{self.url}/file", url, name, user, content_type=content_type
        )

    async def send_image(self, url, name, user, content_type=None):
        return await self.client.send_file(
            f"{self.url}/image", url, name, user, content_type=content_type
        )

    async def delete_file(self, url):
        return await self.client.delete(f"{self.url}/file", {"url": url})

    async def delete_image(self, url):
        return await self.client.delete(f"{self.url}/image", {"url": url})

    async def hide(self, user_id):
        return await self.client.post(f"{self.url}/hide", data={"user_id": user_id})

    async def show(self, user_id):
        return await self.client.post(f"{self.url}/show", data={"user_id": user_id})

    async def mute(self, user_id, expiration=None):
        """
        Mutes the channel for the given user for an optional expiration in milliseconds

        :param user_id: the id of the user muting the channel
        :param expiration: optional expiration in ms. If not given, mute doesn't expire until unmutes
        :return: The server response
        """

        params = {
            "user_id": user_id,
            "channel_cid": self.cid,
        }
        if expiration:
            params["expiration"] = expiration
        return await self.client.post("moderation/mute/channel", data=params)

    async def unmute(self, user_id):
        """
        Unmutes the channel for the given user

        :param user_id: the id of the user unmuting the channel
        :return: The server response
        """

        params = {
            "user_id": user_id,
            "channel_cid": self.cid,
        }
        return await self.client.post("moderation/unmute/channel", data=params)



__author__ = "Tommaso Barbugli"
__copyright__ = "Copyright 2019-2021, Stream.io, Inc"
__version__ = "3.12.0"
__maintainer__ = "Tommaso Barbugli"
__email__ = "support@getstream.io"
__status__ = "Production"



import json


class StreamChannelException(Exception):
    pass


class StreamAPIException(Exception):
    def __init__(self, text, status_code):
        self.response_text = text
        self.status_code = status_code
        self.json_response = False

        try:
            parsed_response = json.loads(text)
            self.error_code = parsed_response.get("code", "unknown")
            self.error_message = parsed_response.get("message", "unknown")
            self.json_response = True
        except ValueError:
            pass

    def __str__(self):
        if self.json_response:
            return f'StreamChat error code {self.error_code}: {self.error_message}"'
        else:
            return f"StreamChat error HTTP code: {self.status_code}"



import json
from urllib.parse import urlparse
from urllib.request import Request, urlopen

import requests
import datetime

from stream_chat.__pkg__ import __version__
from stream_chat.base.client import StreamChatInterface
from stream_chat.base.exceptions import StreamAPIException

from .channel import Channel


def get_user_agent():
    return "stream-python-client-%s" % __version__


def get_default_header():
    base_headers = {
        "Content-type": "application/json",
        "X-Stream-Client": get_user_agent(),
    }
    return base_headers


class StreamChat(StreamChatInterface):
    def __init__(self, api_key, api_secret, timeout=6.0, **options):
        super().__init__(
            api_key=api_key, api_secret=api_secret, timeout=timeout, **options
        )
        self.session = requests.Session()

    def _parse_response(self, response):
        try:
            parsed_result = json.loads(response.text) if response.text else {}
        except ValueError:
            raise StreamAPIException(response.text, response.status_code)
        if response.status_code >= 399:
            raise StreamAPIException(response.text, response.status_code)
        return parsed_result

    def _make_request(self, method, relative_url, params=None, data=None):
        params = params or {}
        data = data or {}
        serialized = None
        default_params = self.get_default_params()
        default_params.update(params)
        headers = get_default_header()
        headers["Authorization"] = self.auth_token
        headers["stream-auth-type"] = "jwt"

        url = f"{self.base_url}/{relative_url}"

        if method.__name__ in ["post", "put", "patch"]:
            serialized = json.dumps(data)

        response = method(
            url,
            data=serialized,
            headers=headers,
            params=default_params,
            timeout=self.timeout,
        )
        return self._parse_response(response)

    def put(self, relative_url, params=None, data=None):
        return self._make_request(self.session.put, relative_url, params, data)

    def post(self, relative_url, params=None, data=None):
        return self._make_request(self.session.post, relative_url, params, data)

    def get(self, relative_url, params=None):
        return self._make_request(self.session.get, relative_url, params, None)

    def delete(self, relative_url, params=None):
        return self._make_request(self.session.delete, relative_url, params, None)

    def patch(self, relative_url, params=None, data=None):
        return self._make_request(self.session.patch, relative_url, params, data)

    def update_app_settings(self, **settings):
        return self.patch("app", data=settings)

    def get_app_settings(self):
        return self.get("app")

    def update_users(self, users):
        return self.post("users", data={"users": {u["id"]: u for u in users}})

    def update_user(self, user):
        return self.update_users([user])

    def update_users_partial(self, updates):
        return self.patch("users", data={"users": updates})

    def update_user_partial(self, update):
        return self.update_users_partial([update])

    def delete_user(self, user_id, **options):
        return self.delete(f"users/{user_id}", options)

    def deactivate_user(self, user_id, **options):
        return self.post(f"users/{user_id}/deactivate", data=options)

    def reactivate_user(self, user_id, **options):
        return self.post(f"users/{user_id}/reactivate", data=options)

    def export_user(self, user_id, **options):
        return self.get(f"users/{user_id}/export", options)

    def ban_user(self, target_id, **options):
        data = {"target_user_id": target_id, **options}
        return self.post("moderation/ban", data=data)

    def unban_user(self, target_id, **options):
        params = {"target_user_id": target_id, **options}
        return self.delete("moderation/ban", params)

    def flag_message(self, target_id, **options):
        data = {"target_message_id": target_id, **options}
        return self.post("moderation/flag", data=data)

    def unflag_message(self, target_id, **options):
        data = {"target_message_id": target_id, **options}
        return self.post("moderation/unflag", data=data)

    def query_message_flags(self, filter_conditions, **options):
        params = {
            **options,
            "filter_conditions": filter_conditions,
        }
        return self.get(
            "moderation/flags/message", params={"payload": json.dumps(params)}
        )

    def flag_user(self, target_id, **options):
        data = {"target_user_id": target_id, **options}
        return self.post("moderation/flag", data=data)

    def unflag_user(self, target_id, **options):
        data = {"target_user_id": target_id, **options}
        return self.post("moderation/unflag", data=data)

    def mute_user(self, target_id, user_id, **options):
        """
        Create a mute

        :param target_id: the user getting muted
        :param user_id: the user muting the target
        :param options: additional mute options
        :return:
        """
        data = {"target_id": target_id, "user_id": user_id, **options}
        return self.post("moderation/mute", data=data)

    def unmute_user(self, target_id, user_id):
        """
        Removes a mute

        :param target_id: the user getting un-muted
        :param user_id: the user muting the target
        :return:
        """

        data = {"target_id": target_id, "user_id": user_id}
        return self.post("moderation/unmute", data=data)

    def mark_all_read(self, user_id):
        return self.post("channels/read", data={"user": {"id": user_id}})

    def update_message(self, message):
        if message.get("id") is None:
            raise ValueError("message must have an id")
        return self.post(f"messages/{message['id']}", data={"message": message})

    def delete_message(self, message_id, **options):
        return self.delete(f"messages/{message_id}", options)

    def get_message(self, message_id):
        return self.get(f"messages/{message_id}")

    def query_users(self, filter_conditions, sort=None, **options):
        params = options.copy()
        params.update(
            {"filter_conditions": filter_conditions, "sort": self.normalize_sort(sort)}
        )
        return self.get("users", params={"payload": json.dumps(params)})

    def query_channels(self, filter_conditions, sort=None, **options):
        params = {"state": True, "watch": False, "presence": False}
        params.update(options)
        params.update(
            {"filter_conditions": filter_conditions, "sort": self.normalize_sort(sort)}
        )
        return self.get("channels", params={"payload": json.dumps(params)})

    def create_channel_type(self, data):
        if "commands" not in data or not data["commands"]:
            data["commands"] = ["all"]
        return self.post("channeltypes", data=data)

    def get_channel_type(self, channel_type):
        return self.get(f"channeltypes/{channel_type}")

    def list_channel_types(self):
        return self.get("channeltypes")

    def update_channel_type(self, channel_type, **settings):
        return self.put(f"channeltypes/{channel_type}", data=settings)

    def delete_channel_type(self, channel_type):
        """
        Delete a type of channel

        :param channel_type: the channel type
        :return:
        """
        return self.delete(f"channeltypes/{channel_type}")

    def channel(self, channel_type, channel_id=None, data=None):
        """
        Creates a channel object

        :param channel_type: the channel type
        :param channel_id: the id of the channel
        :param data: additional data, ie: {"members":[id1, id2, ...]}
        :return: Channel
        """
        return Channel(self, channel_type, channel_id, data)

    def list_commands(self):
        return self.get("commands")

    def create_command(self, data):
        return self.post("commands", data=data)

    def delete_command(self, name):
        return self.delete(f"commands/{name}")

    def get_command(self, name):
        return self.get(f"commands/{name}")

    def update_command(self, name, **settings):
        return self.put(f"commands/{name}", data=settings)

    def add_device(self, device_id, push_provider, user_id):
        """
        Add a device to a user

        :param device_id: the id of the device
        :param push_provider: the push provider used (apn or firebase)
        :param user_id: the id of the user
        :return:
        """
        return self.post(
            "devices",
            data={"id": device_id, "push_provider": push_provider, "user_id": user_id},
        )

    def delete_device(self, device_id, user_id):
        """
        Delete a device for a user

        :param device_id: the id of the device
        :param user_id: the id of the user
        :return:
        """
        return self.delete("devices", {"id": device_id, "user_id": user_id})

    def get_devices(self, user_id):
        """
        Get the list of devices for a user

        :param user_id: the id of the user
        :return: list of devices
        """
        return self.get("devices", {"user_id": user_id})

    def get_rate_limits(
        self, server_side=False, android=False, ios=False, web=False, endpoints=None
    ):
        """
        Get rate limit quotas and usage.
        If no params are toggled, all limits for all endpoints are returned.

        :param server_side: if true, show server_side limits.
        :param android: if true, show android limits.
        :param ios: if true, show ios limits.
        :param web: if true, show web limits.
        :param endpoints: restrict returned limits to the given list of endpoints.
        """
        params = {}
        if server_side:
            params["server_side"] = "true"
        if android:
            params["android"] = "true"
        if ios:
            params["ios"] = "true"
        if web:
            params["web"] = "true"
        if endpoints is not None and len(endpoints) > 0:
            params["endpoints"] = ",".join(endpoints)

        return self.get("rate_limits", params)

    def search(self, filter_conditions, query, sort=None, **options):
        if "offset" in options:
            if sort or "next" in options:
                raise ValueError("cannot use offset with sort or next parameters")
        params = self.create_search_params(filter_conditions, query, sort, **options)
        return self.get("search", params={"payload": json.dumps(params)})

    def send_file(self, uri, url, name, user, content_type=None):
        headers = {
            "Authorization": self.auth_token,
            "stream-auth-type": "jwt",
            "X-Stream-Client": get_user_agent(),
        }
        parts = urlparse(url)
        if parts[0] == "":
            with open(url, "rb") as f:
                content = f.read()
        else:
            content = urlopen(
                Request(url, headers={"User-Agent": "Mozilla/5.0"})
            ).read()
        response = requests.post(
            f"{self.base_url}/{uri}",
            params=self.get_default_params(),
            data={"user": json.dumps(user)},
            files={"file": (name, content, content_type)},
            headers=headers,
        )
        return self._parse_response(response)

    def create_blocklist(self, name, words):
        """
        Create a blocklist

        :param name: the name of the blocklist
        :param words: list of blocked words
        :return:
        """
        return self.post("blocklists", data={"name": name, "words": words})

    def list_blocklists(self):
        """
        List blocklists

        :return: list of blocklists
        """
        return self.get("blocklists")

    def get_blocklist(self, name):
        """Get a blocklist by name

        :param name: the name of the blocklist
        :return: blocklist dict representation
        """
        return self.get(f"blocklists/{name}")

    def update_blocklist(self, name, words):
        """
        Update a blocklist

        :param name: the name of the blocklist
        :param words: the list of blocked words (replaces the current list)
        :return:
        """
        return self.put(f"blocklists/{name}", data={"words": words})

    def delete_blocklist(self, name):
        """Delete a blocklist by name

        :param: the name of the blocklist
        :return:
        """
        return self.delete(f"blocklists/{name}")

    def check_sqs(self, sqs_key=None, sqs_secret=None, sqs_url=None):
        """
        Check SQS Push settings

        When no parameters are given, the current SQS app settings are used

        :param sqs_key: AWS access key
        :param sqs_secret: AWS secret key
        :param sqs_url: URL to SQS queue
        :return:
        """
        data = {"sqs_key": sqs_key, "sqs_secret": sqs_secret, "sqs_url": sqs_url}
        return self.post("check_sqs", data=data)

    def get_permission(self, id):
        """
        Get the definition for a permission

        :param id: ID of the permission
        """
        return self.get(f"permissions/{id}")

    def create_permission(self, permission):
        """
        Create a custom permission

        :param permission: Definition of the permission
        """
        return self.post("permissions", data=permission)

    def update_permission(self, id, permission):
        """
        Update a custom permission

        :param id: ID of the permission
        :param permission: New definition of the permission
        """
        return self.put(f"permissions/{id}", data=permission)

    def delete_permission(self, id):
        """
        Delete a custom permission

        :param id: ID of the permission
        """
        return self.delete(f"permissions/{id}")

    def list_permissions(self):
        """
        List custom permissions of the app
        """
        return self.get("permissions")

    def create_role(self, name):
        """
        Create a custom role

        :param name: Name of the role
        """
        return self.post("roles", data={"name": name})

    def delete_role(self, name):
        """
        Delete a custom role

        :param name: Name of the role
        """
        return self.delete(f"roles/{name}")

    def list_roles(self):
        """
        List all roles of the app
        """
        return self.get("roles")

    def create_segment(self, segment):
        """
        Create a segment
        """
        return self.post("segments", data={"segment": segment})

    def get_segment(self, segment_id):
        """
        Get a segment by id
        """
        return self.get(f"segments/{segment_id}")

    def list_segments(self, **params):
        """
        List segments
        """
        return self.get("segments", params)

    def update_segment(self, segment_id, data):
        """
        Update a segment by id
        """
        return self.put(f"segments/{segment_id}", data=data)

    def delete_segment(self, segment_id):
        """
        Delete a segment by id
        """
        return self.delete(f"segments/{segment_id}")

    def create_campaign(self, campaign):
        """
        Create a campaign
        """
        return self.post("campaigns", data={"campaign": campaign})

    def get_campaign(self, campaign_id):
        """
        Get a campaign by id
        """
        return self.get(f"campaigns/{campaign_id}")

    def list_campaigns(self, **params):
        """
        List campaigns
        """
        return self.get("campaigns", params)

    def update_campaign(self, campaign_id, data):
        """
        Update a campaign
        """
        return self.put(f"campaigns/{campaign_id}", data=data)

    def delete_campaign(self, campaign_id):
        """
        Delete a campaign by id
        """
        return self.delete(f"campaigns/{campaign_id}")

    def schedule_campaign(self, campaign_id, send_at=None):
        """
        Schedule a campaign at given time
        """
        return self.patch(
            f"campaigns/{campaign_id}/schedule", data={"send_at": send_at}
        )

    def stop_campaign(self, campaign_id):
        """
        Stop a in progress campaign
        """
        return self.patch(f"campaigns/{campaign_id}/stop")

    def resume_campaign(self, campaign_id):
        """
        Resume a stopped campaign
        """
        return self.patch(f"campaigns/{campaign_id}/resume")

    def test_campaign(self, campaign_id, users):
        """
        Trigger a test send of the given campaing to given users
        """
        return self.post(f"campaigns/{campaign_id}/test", data={"users": users})

    def revoke_tokens(self, before):
        """
        Revokes tokens for an application
        :param before: date before which the tokens are to be revoked, to reset pass None
        """
        if isinstance(before, datetime.datetime):
            before = before.isoformat()

        self.update_app_settings({"revoke_tokens_issued_before": before})

    def revoke_user_token(self, user_id, before):
        """
        Revokes token for a user
        :param user_id: user_id of user for which the token needs to be revoked
        :param before: date before which the tokens are to be revoked, to reset pass None
        """
        self.revoke_users_token([user_id], before)

    def revoke_users_token(self, user_ids, before):
        """
        Revokes tokens for given users
        :param user_ids: user_ids for user for whom the token needs to be revoked
        :param before: date before which the tokens are to be revoked, to reset pass None
        """
        if isinstance(before, datetime.datetime):
            before = before.isoformat()

        updates = []
        for user_id in user_ids:
            updates.append(
                {"id": user_id, "set": {"revoke_tokens_issued_before": before}}
            )
        self.update_users_partial(updates)

    def export_channel(
        self,
        channel_type,
        channel_id,
        messages_since=None,
        messages_until=None,
    ):
        """
        Requests a channel export
        :param channel_type: channel_type of channel which needs to be exported
        :param channel_id: channel_id of channel which needs to be exported
        :param messages_since: RFC-3339 string or datetime to filter messages since that time, optional
        :param messages_until: RFC-3339 string or datetime to filter messages until that time, optional
        :type channel_type: str
        :type channel_id: str
        :type messages_since: Union[str, datetime.datetime]
        :type messages_until: Union[str, datetime.datetime]
        """
        if isinstance(messages_since, datetime.datetime):
            messages_since = messages_since.isoformat()
        if isinstance(messages_until, datetime.datetime):
            messages_until = messages_until.isoformat()

        return self.export_channels(
            [
                {
                    "id": channel_id,
                    "type": channel_type,
                    "messages_since": messages_since,
                    "messages_until": messages_until,
                }
            ]
        )

    def export_channels(self, channels):
        """
        Requests a channels export
        :param channels: list of channel's data which need to be exported with keys:
        - `channel_type`: str
        - `channel_id`: str
        - `messages_since` (optional, nullable): str
        - `messages_until` (optional, nullable): str
        :type channels: List[Dict[str, str]]
        """
        return self.post("export_channels", data={"channels": channels})

    def get_export_channel_status(self, task_id):
        """
        Retrieves status of export
        :param task_id: task_id of task which status needs to be retrieved
        :type task_id: str
        """
        return self.get(f"export_channels/{task_id}")



import abc

from stream_chat.base.exceptions import StreamChannelException


class ChannelInterface(abc.ABC):
    def __init__(self, client, channel_type, channel_id=None, custom_data=None):
        self.channel_type = channel_type
        self.id = channel_id
        self.client = client
        self.custom_data = custom_data
        if self.custom_data is None:
            self.custom_data = {}

    @property
    def url(self):
        if self.id is None:
            raise StreamChannelException("channel does not have an id")
        return f"channels/{self.channel_type}/{self.id}"

    @property
    def cid(self):
        if self.id is None:
            raise StreamChannelException("channel does not have an id")
        return f"{self.channel_type}:{self.id}"

    @abc.abstractmethod
    def send_message(self, message, user_id, **options):
        """
        Send a message to this channel

        :param message: the Message object
        :param user_id: the ID of the user that created the message
        :return: the Server Response
        """
        pass

    @abc.abstractmethod
    def send_event(self, event, user_id):
        """
        Send an event on this channel

        :param event: event data, ie {type: 'message.read'}
        :param user_id: the ID of the user sending the event
        :return: the Server Response
        """
        pass

    @abc.abstractmethod
    def send_reaction(self, message_id, reaction, user_id):
        """
        Send a reaction about a message

        :param message_id: the message id
        :param reaction: the reaction object, ie {type: 'love'}
        :param user_id: the ID of the user that created the reaction
        :return: the Server Response
        """
        pass

    @abc.abstractmethod
    def delete_reaction(self, message_id, reaction_type, user_id):
        """
        Delete a reaction by user and type

        :param message_id: the id of the message from which te remove the reaction
        :param reaction_type: the type of reaction that should be removed
        :param user_id: the id of the user
        :return: the Server Response
        """
        pass

    @abc.abstractmethod
    def create(self, user_id):
        """
        Create the channel

        :param user_id: the ID of the user creating this channel
        :return:
        """
        pass

    @abc.abstractmethod
    def query(self, **options):
        """
        Query the API for this channel, get messages, members or other channel fields

        :param options: the query options, check docs on https://getstream.io/chat/docs/
        :return: Returns a query response
        """
        pass

    @abc.abstractmethod
    def query_members(self, filter_conditions, sort=None, **options):
        """
        Query the API for this channel to filter, sort and paginate its members efficiently.

        :param filter_conditions: filters, checks docs on https://getstream.io/chat/docs/
        :param sort: sorting field and direction slice, check docs on https://getstream.io/chat/docs/
        :param options: pagination or members based channel searching details
        :return: Returns members response

        eg.
        channel.query_members(filter_conditions={"name": "tommaso"},
                              sort=[{"field": "created_at", "direction": -1}],
                              offset=0,
                              limit=10)
        """
        pass

    @abc.abstractmethod
    def update(self, channel_data, update_message=None):
        """
        Edit the channel's custom properties

        :param channel_data: the object to update the custom properties of this channel with
        :param update_message: optional update message
        :return: The server response
        """
        pass

    @abc.abstractmethod
    def update_partial(self, to_set=None, to_unset=None):
        """
        Update channel partially

        :param to_set: a dictionary of key/value pairs to set or to override
        :param to_unset: a list of keys to clear
        """
        pass

    @abc.abstractmethod
    def delete(self):
        """
        Delete the channel. Messages are permanently removed.

        :return: The server response
        """
        pass

    @abc.abstractmethod
    def truncate(self):
        """
        Removes all messages from the channel

        :return: The server response
        """
        pass

    @abc.abstractmethod
    def add_members(self, user_ids):
        """
        Adds members to the channel

        :param user_ids: user IDs to add as members
        :return:
        """
        pass

    @abc.abstractmethod
    def invite_members(self, user_ids):
        """
        invite members to the channel

        :param user_ids: user IDs to invite
        :return:
        """
        pass

    @abc.abstractmethod
    def add_moderators(self, user_ids):
        """
        Adds moderators to the channel

        :param user_ids: user IDs to add as moderators
        :return:
        """
        pass

    @abc.abstractmethod
    def remove_members(self, user_ids):
        """
        Remove members from the channel

        :param user_ids: user IDs to remove from the member list
        :return:
        """
        pass

    @abc.abstractmethod
    def demote_moderators(self, user_ids):
        """
        Demotes moderators from the channel

        :param user_ids: user IDs to demote
        :return:
        """
        pass

    @abc.abstractmethod
    def mark_read(self, user_id, **data):
        """
        Send the mark read event for this user, only works if the `read_events` setting is enabled

        :param user_id: the user ID for the event
        :param data: additional data, ie {"message_id": last_message_id}
        :return: The server response
        """
        pass

    @abc.abstractmethod
    def get_replies(self, parent_id, **options):
        """
        List the message replies for a parent message

        :param parent_id: The message parent id, ie the top of the thread
        :param options: Pagination params, ie {limit:10, id_lte: 10}
        :return: A response with a list of messages
        """
        pass

    @abc.abstractmethod
    def get_reactions(self, message_id, **options):
        """
        List the reactions, supports pagination

        :param message_id: The message id
        :param options: Pagination params, ie {"limit":10, "id_lte": 10}
        :return: A response with a list of reactions
        """
        pass

    @abc.abstractmethod
    def ban_user(self, target_id, **options):
        """
        Bans a user from this channel

        :param target_id: the ID of the user to ban
        :param options: additional ban options, ie {"timeout": 3600, "reason": "offensive language is not allowed here"}
        :return: The server response
        """
        pass

    @abc.abstractmethod
    def unban_user(self, target_id, **options):
        """
        Removes the ban for a user on this channel

        :param target_id: the ID of the user to unban
        :return: The server response
        """
        pass

    @abc.abstractmethod
    def accept_invite(self, user_id, **data):
        pass

    @abc.abstractmethod
    def reject_invite(self, user_id, **data):
        pass

    @abc.abstractmethod
    def send_file(self, url, name, user, content_type=None):
        pass

    @abc.abstractmethod
    def send_image(self, url, name, user, content_type=None):
        pass

    @abc.abstractmethod
    def delete_file(self, url):
        pass

    @abc.abstractmethod
    def delete_image(self, url):
        pass

    @abc.abstractmethod
    def hide(self, user_id):
        pass

    @abc.abstractmethod
    def show(self, user_id):
        pass

    @abc.abstractmethod
    def mute(self, user_id, expiration=None):
        pass

    @abc.abstractmethod
    def unmute(self, user_id):
        pass


def add_user_id(payload, user_id):
    return {**payload, "user": {"id": user_id}}



import abc
import collections
import hashlib
import hmac

import jwt


class StreamChatInterface(abc.ABC):
    session = None

    @abc.abstractmethod
    def __init__(self, api_key, api_secret, timeout=6.0, **options):
        self.api_key = api_key
        self.api_secret = api_secret
        self.timeout = timeout
        self.options = options
        self.base_url = options.get("base_url", "https://chat.stream-io-api.com")
        self.auth_token = jwt.encode(
            {"server": True}, self.api_secret, algorithm="HS256"
        )

    def get_default_params(self):
        return {"api_key": self.api_key}

    def normalize_sort(self, sort=None):
        sort_fields = []
        if isinstance(sort, collections.abc.Mapping):
            sort = [sort]
        if isinstance(sort, list):
            for item in sort:
                if "field" in item and "direction" in item:
                    sort_fields.append(item)
                else:
                    for k, v in item.items():
                        sort_fields.append({"field": k, "direction": v})
        return sort_fields

    def create_token(self, user_id, exp=None, iat=None, **claims):
        payload = {**claims, "user_id": user_id}
        if exp is not None:
            payload["exp"] = exp
        if iat is not None:
            payload["iat"] = iat
        return jwt.encode(payload, self.api_secret, algorithm="HS256")

    def create_search_params(self, filter_conditions, query, sort, **options):
        params = options.copy()
        if isinstance(query, str):
            params.update({"query": query})
        else:
            params.update({"message_filter_conditions": query})
        params.update(
            {"filter_conditions": filter_conditions, "sort": self.normalize_sort(sort)}
        )
        return params

    def verify_webhook(self, request_body, x_signature):
        """
        Verify the signature added to a webhook event

        :param request_body: the request body received from webhook
        :param x_signature: the x-signature header included in the request
        :return: bool
        """
        signature = hmac.new(
            key=self.api_secret.encode(), msg=request_body, digestmod=hashlib.sha256
        ).hexdigest()
        return signature == x_signature

    @abc.abstractmethod
    def update_app_settings(self, **settings):
        pass

    @abc.abstractmethod
    def get_app_settings(self):
        pass

    @abc.abstractmethod
    def update_users(self, users):
        pass

    @abc.abstractmethod
    def update_user(self, user):
        pass

    @abc.abstractmethod
    def update_users_partial(self, updates):
        pass

    @abc.abstractmethod
    def update_user_partial(self, update):
        pass

    @abc.abstractmethod
    def delete_user(self, user_id, **options):
        pass

    @abc.abstractmethod
    def deactivate_user(self, user_id, **options):
        pass

    @abc.abstractmethod
    def reactivate_user(self, user_id, **options):
        pass

    @abc.abstractmethod
    def export_user(self, user_id, **options):
        pass

    @abc.abstractmethod
    def ban_user(self, target_id, **options):
        pass

    @abc.abstractmethod
    def unban_user(self, target_id, **options):
        pass

    @abc.abstractmethod
    def flag_message(self, target_id, **options):
        pass

    @abc.abstractmethod
    def unflag_message(self, target_id, **options):
        pass

    @abc.abstractmethod
    def query_message_flags(self, filter_conditions, **options):
        pass

    @abc.abstractmethod
    def flag_user(self, target_id, **options):
        pass

    @abc.abstractmethod
    def unflag_user(self, target_id, **options):
        pass

    @abc.abstractmethod
    def mute_user(self, target_id, user_id, **options):
        """
        Create a mute

        :param target_id: the user getting muted
        :param user_id: the user muting the target
        :param options: additional mute options
        :return:
        """
        pass

    @abc.abstractmethod
    def unmute_user(self, target_id, user_id):
        """
        Removes a mute

        :param target_id: the user getting un-muted
        :param user_id: the user muting the target
        :return:
        """

        pass

    @abc.abstractmethod
    def mark_all_read(self, user_id):
        pass

    @abc.abstractmethod
    def update_message(self, message):
        pass

    @abc.abstractmethod
    def delete_message(self, message_id, **options):
        pass

    @abc.abstractmethod
    def get_message(self, message_id):
        pass

    @abc.abstractmethod
    def query_users(self, filter_conditions, sort=None, **options):
        pass

    @abc.abstractmethod
    def query_channels(self, filter_conditions, sort=None, **options):
        pass

    @abc.abstractmethod
    def create_channel_type(self, data):
        pass

    @abc.abstractmethod
    def get_channel_type(self, channel_type):
        pass

    @abc.abstractmethod
    def list_channel_types(self):
        pass

    @abc.abstractmethod
    def update_channel_type(self, channel_type, **settings):
        pass

    @abc.abstractmethod
    def delete_channel_type(self, channel_type):
        """
        Delete a type of channel

        :param channel_type: the channel type
        :return:
        """
        pass

    @abc.abstractmethod
    def channel(self, channel_type, channel_id=None, data=None):
        """
        Creates a channel object

        :param channel_type: the channel type
        :param channel_id: the id of the channel
        :param data: additional data, ie: {"members":[id1, id2, ...]}
        :return: Channel
        """
        pass

    @abc.abstractmethod
    def list_commands(self):
        pass

    @abc.abstractmethod
    def create_command(self, data):
        pass

    @abc.abstractmethod
    def delete_command(self, name):
        pass

    @abc.abstractmethod
    def get_command(self, name):
        pass

    @abc.abstractmethod
    def update_command(self, name, **settings):
        pass

    @abc.abstractmethod
    def add_device(self, device_id, push_provider, user_id):
        """
        Add a device to a user

        :param device_id: the id of the device
        :param push_provider: the push provider used (apn or firebase)
        :param user_id: the id of the user
        :return:
        """
        pass

    @abc.abstractmethod
    def delete_device(self, device_id, user_id):
        """
        Delete a device for a user

        :param device_id: the id of the device
        :param user_id: the id of the user
        :return:
        """
        pass

    @abc.abstractmethod
    def get_devices(self, user_id):
        """
        Get the list of devices for a user

        :param user_id: the id of the user
        :return: list of devices
        """
        pass

    @abc.abstractmethod
    def get_rate_limits(self, server_side, android, ios, web, endpoints):
        """
        Get rate limit quotas and usage.
        If no params are toggled, all limits for all endpoints are returned.

        :param server_side: if true, show server_side limits.
        :param android: if true, show android limits.
        :param ios: if true, show ios limits.
        :param web: if true, show web limits.
        :param endpoints: restrict returned limits to the given list of endpoints.
        """
        pass

    @abc.abstractmethod
    def search(self, filter_conditions, query, **options):
        pass

    @abc.abstractmethod
    def send_file(self, uri, url, name, user, content_type=None):
        pass

    @abc.abstractmethod
    def create_blocklist(self, name, words):
        """
        Create a blocklist

        :param name: the name of the blocklist
        :param words: list of blocked words
        :return:
        """
        pass

    @abc.abstractmethod
    def list_blocklists(self):
        """
        List blocklists

        :return: list of blocklists
        """
        pass

    @abc.abstractmethod
    def get_blocklist(self, name):
        """Get a blocklist by name

        :param name: the name of the blocklist
        :return: blocklist dict representation
        """
        pass

    @abc.abstractmethod
    def update_blocklist(self, name, words):
        """
        Update a blocklist

        :param name: the name of the blocklist
        :param words: the list of blocked words (replaces the current list)
        :return:
        """
        pass

    @abc.abstractmethod
    def delete_blocklist(self, name):
        """Delete a blocklist by name

        :param: the name of the blocklist
        :return:
        """
        pass

    @abc.abstractmethod
    def check_sqs(self, sqs_key=None, sqs_secret=None, sqs_url=None):
        """
        Check SQS Push settings

        When no parameters are given, the current SQS app settings are used

        :param sqs_key: AWS access key
        :param sqs_secret: AWS secret key
        :param sqs_url: URL to SQS queue
        :return:
        """
        pass

    @abc.abstractmethod
    def get_permission(self, name):
        """
        Get the definition for a permission

        :param name: Name of the permission
        """
        pass

    @abc.abstractmethod
    def create_permission(self, permission):
        """
        Create a custom permission

        :param permission: Definition of the permission
        """
        pass

    @abc.abstractmethod
    def update_permission(self, name, permission):
        """
        Update a custom permission

        :param name: Name of the permission
        :param permission: New definition of the permission
        """
        pass

    @abc.abstractmethod
    def delete_permission(self, name):
        """
        Delete a custom permission

        :param name: Name of the permission
        """
        pass

    @abc.abstractmethod
    def list_permissions(self):
        """
        List all permissions of the app
        """
        pass

    @abc.abstractmethod
    def create_role(self, name):
        """
        Create a custom role

        :param name: Name of the role
        """
        pass

    @abc.abstractmethod
    def delete_role(self, name):
        """
        Delete a custom role

        :param name: Name of the role
        """
        pass

    @abc.abstractmethod
    def list_roles(self):
        """
        List all roles of the app
        """
        pass

    @abc.abstractmethod
    def create_segment(self, segment):
        """
        Create a segment
        """
        pass

    @abc.abstractmethod
    def get_segment(self, segment_id):
        """
        Get a segment by id
        """
        pass

    @abc.abstractmethod
    def list_segments(self, **params):
        """
        List segments
        """
        pass

    @abc.abstractmethod
    def update_segment(self, segment_id, data):
        """
        Update a segment by id
        """
        pass

    @abc.abstractmethod
    def delete_segment(self, segment_id):
        """
        Delete a segment by id
        """
        pass

    @abc.abstractmethod
    def create_campaign(self, campaign):
        """
        Create a campaign
        """
        pass

    @abc.abstractmethod
    def get_campaign(self, campaign_id):
        """
        Get a campaign by id
        """
        pass

    @abc.abstractmethod
    def list_campaigns(self, **params):
        """
        List campaigns
        """
        pass

    @abc.abstractmethod
    def update_campaign(self, campaign_id, data):
        """
        Update a campaign
        """
        pass

    @abc.abstractmethod
    def delete_campaign(self, campaign_id):
        """
        Delete a campaign by id
        """
        pass

    @abc.abstractmethod
    def schedule_campaign(self, campaign_id, send_at=None):
        """
        Schedule a campaign at given time
        """
        pass

    @abc.abstractmethod
    async def stop_campaign(self, campaign_id):
        """
        Stop a in progress campaign
        """
        pass

    @abc.abstractmethod
    def resume_campaign(self, campaign_id):
        """
        Resume a stopped campaign
        """
        pass

    @abc.abstractmethod
    def test_campaign(self, campaign_id, users):
        """
        Trigger a test send of the given campaing to given users
        """
        pass

    @abc.abstractmethod
    def revoke_tokens(self, since):
        """
        Revoke tokens for an application issued since
        """
        pass

    @abc.abstractmethod
    def revoke_user_token(self, user_id, since):
        """
        Revoke tokens for a user issued since
        """
        pass

    @abc.abstractmethod
    def revoke_users_token(self, user_ids, since):
        """
        Revoke tokens for users issued since
        """
        pass

    @abc.abstractmethod
    def export_channel(
        self, channel_type, channel_id, messages_since=None, messages_until=None
    ):
        """
        Requests a channel export
        """
        pass

    @abc.abstractmethod
    def export_channels(self, channels):
        """
        Requests a channels export
        """
        pass

    @abc.abstractmethod
    def get_export_channel_status(self, task_id):
        """
        Retrieves status of export
        """
        pass



import os
import uuid

import pytest

from stream_chat import StreamChat


def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        if call.excinfo is not None:
            parent = item.parent
            parent._previousfailed = item


def pytest_runtest_setup(item):
    if "incremental" in item.keywords:
        previousfailed = getattr(item.parent, "_previousfailed", None)
        if previousfailed is not None:
            pytest.xfail("previous test failed (%s)" % previousfailed.name)


def pytest_configure(config):
    config.addinivalue_line("markers", "incremental: mark test incremental")


@pytest.fixture(scope="module")
def client():
    base_url = os.environ.get("STREAM_HOST")
    options = {"base_url": base_url} if base_url else {}
    return StreamChat(
        api_key=os.environ["STREAM_KEY"],
        api_secret=os.environ["STREAM_SECRET"],
        timeout=10,
        **options,
    )


@pytest.fixture(scope="function")
def random_user(client):
    user = {"id": str(uuid.uuid4())}
    response = client.update_user(user)
    assert "users" in response
    assert user["id"] in response["users"]
    return user


@pytest.fixture(scope="function")
def server_user(client):
    user = {"id": str(uuid.uuid4())}
    response = client.update_user(user)
    assert "users" in response
    assert user["id"] in response["users"]
    return user


@pytest.fixture(scope="function")
def random_users(client):
    user1 = {"id": str(uuid.uuid4())}
    user2 = {"id": str(uuid.uuid4())}
    client.update_users([user1, user2])
    return [user1, user2]


@pytest.fixture(scope="function")
def channel(client, random_user):
    channel = client.channel(
        "messaging", str(uuid.uuid4()), {"test": True, "language": "python"}
    )
    channel.create(random_user["id"])
    return channel


@pytest.fixture(scope="function")
def command(client):
    response = client.create_command(
        dict(name=str(uuid.uuid4()), description="My command")
    )

    yield response["command"]

    client.delete_command(response["command"]["name"])


@pytest.fixture(scope="module")
def fellowship_of_the_ring(client):
    members = [
        {"id": "frodo-baggins", "name": "Frodo Baggins", "race": "Hobbit", "age": 50},
        {"id": "sam-gamgee", "name": "Samwise Gamgee", "race": "Hobbit", "age": 38},
        {"id": "gandalf", "name": "Gandalf the Grey", "race": "Istari"},
        {"id": "legolas", "name": "Legolas", "race": "Elf", "age": 500},
        {"id": "gimli", "name": "Gimli", "race": "Dwarf", "age": 139},
        {"id": "aragorn", "name": "Aragorn", "race": "Man", "age": 87},
        {"id": "boromir", "name": "Boromir", "race": "Man", "age": 40},
        {
            "id": "meriadoc-brandybuck",
            "name": "Meriadoc Brandybuck",
            "race": "Hobbit",
            "age": 36,
        },
        {"id": "peregrin-took", "name": "Peregrin Took", "race": "Hobbit", "age": 28},
    ]
    client.update_users(members)
    channel = client.channel(
        "team", "fellowship-of-the-ring", {"members": [m["id"] for m in members]}
    )
    channel.create("gandalf")



import sys
from operator import itemgetter
from contextlib import suppress

import jwt
import pytest
import time
import uuid
from stream_chat import StreamChat
from stream_chat.base.exceptions import StreamAPIException


class TestClient(object):
    def test_normalize_sort(self, client):
        expected = [
            {"field": "field1", "direction": 1},
            {"field": "field2", "direction": -1},
        ]
        actual = client.normalize_sort([{"field1": 1}, {"field2": -1}])
        assert actual == expected
        actual = client.normalize_sort(
            [{"field": "field1", "direction": 1}, {"field": "field2", "direction": -1}]
        )
        assert actual == expected
        actual = client.normalize_sort({"field1": 1})
        assert actual == [{"field": "field1", "direction": 1}]
        # The following example is not recommended because the order of the fields is not guaranteed in Python < 3.7
        actual = client.normalize_sort({"field1": 1, "field2": -1})
        if sys.version_info >= (3, 7):
            assert actual == expected
        else:
            # Compare elements regardless of the order
            assert sorted(actual, key=itemgetter("field")) == expected

    def test_mute_user(self, client, random_users):
        response = client.mute_user(random_users[0]["id"], random_users[1]["id"])
        assert "mute" in response
        assert "expires" not in response["mute"]
        assert response["mute"]["target"]["id"] == random_users[0]["id"]
        assert response["mute"]["user"]["id"] == random_users[1]["id"]
        client.unmute_user(random_users[0]["id"], random_users[1]["id"])

    def test_mute_user_with_timeout(self, client, random_users):
        response = client.mute_user(
            random_users[0]["id"], random_users[1]["id"], timeout=10
        )
        assert "mute" in response
        assert "expires" in response["mute"]
        assert response["mute"]["target"]["id"] == random_users[0]["id"]
        assert response["mute"]["user"]["id"] == random_users[1]["id"]
        client.unmute_user(random_users[0]["id"], random_users[1]["id"])

    def test_get_message(self, client, channel, random_user):
        msg_id = str(uuid.uuid4())
        channel.send_message({"id": msg_id, "text": "helloworld"}, random_user["id"])
        client.delete_message(msg_id)
        msg_id = str(uuid.uuid4())
        channel.send_message({"id": msg_id, "text": "helloworld"}, random_user["id"])
        message = client.get_message(msg_id)
        assert message["message"]["id"] == msg_id

    def test_auth_exception(self):
        client = StreamChat(api_key="bad", api_secret="guy")
        with pytest.raises(StreamAPIException):
            client.get_channel_type("team")

    def test_get_channel_types(self, client):
        response = client.get_channel_type("team")
        assert "permissions" in response

    def test_list_channel_types(self, client):
        response = client.list_channel_types()
        assert "channel_types" in response

    def test_update_channel_type(self, client):
        response = client.update_channel_type("team", commands=["ban", "unban"])
        assert "commands" in response
        assert response["commands"] == ["ban", "unban"]

    def test_get_command(self, client, command):
        response = client.get_command(command["name"])
        assert command["name"] == response["name"]

    def test_update_command(self, client, command):
        response = client.update_command(command["name"], description="My new command")
        assert "command" in response
        assert "My new command" == response["command"]["description"]

    def test_list_commands(self, client):
        response = client.list_commands()
        assert "commands" in response

    def test_create_token(self, client):
        token = client.create_token("tommaso")
        assert type(token) is str
        payload = jwt.decode(token, client.api_secret, algorithms=["HS256"])
        assert payload.get("user_id") == "tommaso"

    def test_get_app_settings(self, client):
        configs = client.get_app_settings()
        assert "app" in configs

    def test_update_user(self, client):
        user = {"id": str(uuid.uuid4())}
        response = client.update_user(user)
        assert "users" in response
        assert user["id"] in response["users"]

    def test_update_users(self, client):
        user = {"id": str(uuid.uuid4())}
        response = client.update_users([user])
        assert "users" in response
        assert user["id"] in response["users"]

    def test_update_user_partial(self, client):
        user_id = str(uuid.uuid4())
        client.update_user({"id": user_id, "field": "value"})

        response = client.update_user_partial(
            {"id": user_id, "set": {"field": "updated"}}
        )

        assert "users" in response
        assert user_id in response["users"]
        assert response["users"][user_id]["field"] == "updated"

    def test_delete_user(self, client, random_user):
        response = client.delete_user(random_user["id"])
        assert "user" in response
        assert random_user["id"] == response["user"]["id"]

    def test_deactivate_user(self, client, random_user):
        response = client.deactivate_user(random_user["id"])
        assert "user" in response
        assert random_user["id"] == response["user"]["id"]

    def test_reactivate_user(self, client, random_user):
        response = client.deactivate_user(random_user["id"])
        assert "user" in response
        assert random_user["id"] == response["user"]["id"]
        response = client.reactivate_user(random_user["id"])
        assert "user" in response
        assert random_user["id"] == response["user"]["id"]

    def test_export_user(self, client, fellowship_of_the_ring):
        response = client.export_user("gandalf")
        assert "user" in response
        assert response["user"]["name"] == "Gandalf the Grey"

    def test_ban_user(self, client, random_user, server_user):
        client.ban_user(random_user["id"], user_id=server_user["id"])

    def test_unban_user(self, client, random_user, server_user):
        client.ban_user(random_user["id"], user_id=server_user["id"])
        client.unban_user(random_user["id"], user_id=server_user["id"])

    def test_flag_user(self, client, random_user, server_user):
        client.flag_user(random_user["id"], user_id=server_user["id"])

    def test_unflag_user(self, client, random_user, server_user):
        client.flag_user(random_user["id"], user_id=server_user["id"])
        client.unflag_user(random_user["id"], user_id=server_user["id"])

    def test_mark_all_read(self, client, random_user):
        client.mark_all_read(random_user["id"])

    def test_update_message(self, client, channel, random_user):
        msg_id = str(uuid.uuid4())
        response = channel.send_message(
            {"id": msg_id, "text": "hello world"}, random_user["id"]
        )
        assert response["message"]["text"] == "hello world"
        client.update_message(
            {
                "id": msg_id,
                "awesome": True,
                "text": "helloworld",
                "user": {"id": response["message"]["user"]["id"]},
            }
        )

    def test_delete_message(self, client, channel, random_user):
        msg_id = str(uuid.uuid4())
        channel.send_message({"id": msg_id, "text": "helloworld"}, random_user["id"])
        client.delete_message(msg_id)
        msg_id = str(uuid.uuid4())
        channel.send_message({"id": msg_id, "text": "helloworld"}, random_user["id"])
        client.delete_message(msg_id, hard=True)

    def test_flag_message(self, client, channel, random_user, server_user):
        msg_id = str(uuid.uuid4())
        channel.send_message({"id": msg_id, "text": "helloworld"}, random_user["id"])
        client.flag_message(msg_id, user_id=server_user["id"])

    def test_query_message_flags(self, client, channel, random_user, server_user):
        msg_id = str(uuid.uuid4())
        channel.send_message({"id": msg_id, "text": "helloworld"}, random_user["id"])
        client.flag_message(msg_id, user_id=server_user["id"])
        response = client.query_message_flags({"channel_cid": channel.cid})
        assert len(response["flags"]) == 1
        response = client.query_message_flags({"user_id": {"$in": [random_user["id"]]}})
        assert len(response["flags"]) == 1

    def test_unflag_message(self, client, channel, random_user, server_user):
        msg_id = str(uuid.uuid4())
        channel.send_message({"id": msg_id, "text": "helloworld"}, random_user["id"])
        client.flag_message(msg_id, user_id=server_user["id"])
        client.unflag_message(msg_id, user_id=server_user["id"])

    def test_query_users_young_hobbits(self, client, fellowship_of_the_ring):
        response = client.query_users({"race": {"$eq": "Hobbit"}}, {"age": -1})
        assert len(response["users"]) == 4
        assert [50, 38, 36, 28] == [u["age"] for u in response["users"]]

    def test_devices(self, client, random_user):
        response = client.get_devices(random_user["id"])
        assert "devices" in response
        assert len(response["devices"]) == 0

        client.add_device(str(uuid.uuid4()), "apn", random_user["id"])
        response = client.get_devices(random_user["id"])
        assert len(response["devices"]) == 1

        client.delete_device(response["devices"][0]["id"], random_user["id"])
        client.add_device(str(uuid.uuid4()), "apn", random_user["id"])
        response = client.get_devices(random_user["id"])
        assert len(response["devices"]) == 1

    def test_get_rate_limits(self, client):
        response = client.get_rate_limits()
        assert "server_side" in response
        assert "android" in response
        assert "ios" in response
        assert "web" in response

        response = client.get_rate_limits(server_side=True, android=True)
        assert "server_side" in response
        assert "android" in response
        assert "ios" not in response
        assert "web" not in response

        response = client.get_rate_limits(
            server_side=True, android=True, endpoints=["GetRateLimits", "SendMessage"]
        )
        assert "server_side" in response
        assert "android" in response
        assert "ios" not in response
        assert "web" not in response
        assert len(response["android"]) == 2
        assert len(response["server_side"]) == 2
        assert (
            response["android"]["GetRateLimits"]["limit"]
            == response["android"]["GetRateLimits"]["remaining"]
        )
        assert (
            response["server_side"]["GetRateLimits"]["limit"]
            > response["server_side"]["GetRateLimits"]["remaining"]
        )

    @pytest.mark.xfail
    def test_search_with_sort(self, client, channel, random_user):
        text = str(uuid.uuid4())
        ids = ["0" + text, "1" + text]
        channel.send_message(
            {"text": text, "id": ids[0]},
            random_user["id"],
        )
        channel.send_message(
            {"text": text, "id": ids[1]},
            random_user["id"],
        )
        response = client.search(
            {"type": "messaging"}, text, **{"limit": 1, "sort": [{"created_at": -1}]}
        )
        # searches all channels so make sure at least one is found
        assert len(response["results"]) >= 1
        assert response["next"] is not None
        assert ids[1] == response["results"][0]["message"]["id"]
        response = client.search(
            {"type": "messaging"}, text, **{"limit": 1, "next": response["next"]}
        )
        assert len(response["results"]) >= 1
        assert response["previous"] is not None
        assert response["next"] is None
        assert ids[0] == response["results"][0]["message"]["id"]

    def test_search(self, client, channel, random_user):
        query = "supercalifragilisticexpialidocious"
        channel.send_message(
            {"text": f"How many syllables are there in {query}?"},
            random_user["id"],
        )
        channel.send_message(
            {"text": "Does 'cious' count as one or two?"}, random_user["id"]
        )
        response = client.search(
            {"type": "messaging"}, query, **{"limit": 2, "offset": 0}
        )
        # searches all channels so make sure at least one is found
        assert len(response["results"]) >= 1
        assert query in response["results"][0]["message"]["text"]
        response = client.search(
            {"type": "messaging"}, "cious", **{"limit": 12, "offset": 0}
        )
        for message in response["results"]:
            assert query not in message["message"]["text"]

    def test_search_message_filters(self, client, channel, random_user):
        query = "supercalifragilisticexpialidocious"
        channel.send_message(
            {"text": f"How many syllables are there in {query}?"},
            random_user["id"],
        )
        channel.send_message(
            {"text": "Does 'cious' count as one or two?"}, random_user["id"]
        )
        response = client.search(
            {"type": "messaging"},
            {"text": {"$q": query}},
            **{
                "limit": 2,
                "offset": 0,
            },
        )
        assert len(response["results"]) >= 1
        assert query in response["results"][0]["message"]["text"]

    def test_search_offset_with_sort(self, client):
        query = "supercalifragilisticexpialidocious"
        with pytest.raises(ValueError):
            client.search(
                {"type": "messaging"},
                query,
                **{"limit": 2, "offset": 1, "sort": [{"created_at": -1}]},
            )

    def test_search_offset_with_next(self, client):
        query = "supercalifragilisticexpialidocious"
        with pytest.raises(ValueError):
            client.search(
                {"type": "messaging"}, query, **{"limit": 2, "offset": 1, "next": query}
            )

    def test_query_channels_members_in(self, client, fellowship_of_the_ring):
        response = client.query_channels({"members": {"$in": ["gimli"]}}, {"id": 1})
        assert len(response["channels"]) == 1
        assert response["channels"][0]["channel"]["id"] == "fellowship-of-the-ring"
        assert len(response["channels"][0]["members"]) == 9

    def test_create_blocklist(self, client):
        client.create_blocklist(name="Foo", words=["fudge", "heck"])

    def test_list_blocklists(self, client):
        response = client.list_blocklists()
        assert len(response["blocklists"]) == 2
        blocklist_names = {blocklist["name"] for blocklist in response["blocklists"]}
        assert "Foo" in blocklist_names

    def test_get_blocklist(self, client):
        response = client.get_blocklist("Foo")
        assert response["blocklist"]["name"] == "Foo"
        assert response["blocklist"]["words"] == ["fudge", "heck"]

    def test_update_blocklist(self, client):
        client.update_blocklist("Foo", words=["dang"])
        response = client.get_blocklist("Foo")
        assert response["blocklist"]["words"] == ["dang"]

    def test_delete_blocklist(self, client):
        client.delete_blocklist("Foo")

    def test_check_sqs(self, client):
        response = client.check_sqs("key", "secret", "https://foo.com/bar")
        assert response["status"] == "error"
        assert "invalid SQS url" in response["error"]

    @pytest.mark.skip(reason="slow and flaky due to waits")
    def test_custom_permission_and_roles(self, client):
        id, role = "my-custom-permission", "god"

        def wait():
            time.sleep(3)

        with suppress(Exception):
            client.delete_permission(id)
            wait()
        with suppress(Exception):
            client.delete_role(role)
            wait()

        custom = {
            "id": id,
            "name": "My Custom Permission",
            "action": "DeleteChannel",
            "owner": False,
            "same_team": True,
        }

        client.create_permission(custom)
        wait()
        response = client.get_permission(id)
        assert response["permission"]["id"] == id
        assert response["permission"]["custom"]
        assert not response["permission"]["owner"]
        assert response["permission"]["action"] == custom["action"]

        custom["owner"] = True
        client.update_permission(id, custom)

        wait()
        response = client.get_permission(id)
        assert response["permission"]["id"] == id
        assert response["permission"]["custom"]
        assert response["permission"]["owner"]
        assert response["permission"]["action"] == custom["action"]

        response = client.list_permissions()
        original_len = len(response["permissions"])
        assert response["permissions"][0]["id"] == id
        client.delete_permission(id)
        wait()
        response = client.list_permissions()
        assert len(response["permissions"]) == original_len - 1

        client.create_role(role)
        wait()
        response = client.list_roles()
        assert role in response["roles"]
        client.delete_role(role)
        wait()
        response = client.list_roles()
        assert role not in response["roles"]



import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

install_requires = [
    "pycryptodomex>=3.8.1,<4",
    "requests>=2.22.0,<3",
    "chardet>=2.0.0,<4",
    "aiodns>=2.0.0",
    "aiohttp>=3.6.0,<4",
    "aiofile>=3.1,<4",
    "pyjwt>=2.0.0,<3",
]
long_description = open("README.md", "r").read()
tests_require = ["pytest", "pytest-asyncio"]
ci_require = ["black", "flake8", "pytest-cov"]

about = {}
with open("stream_chat/__pkg__.py") as fp:
    exec(fp.read(), about)


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        pytest_cmd = ["stream_chat/", "-v"]

        try:
            pytest_cmd += [
                "--cov=stream_chat/",
                "--cov-report=xml",
            ]
        except ImportError:
            pass

        errno = pytest.main(pytest_cmd)
        sys.exit(errno)


setup(
    name="stream-chat",
    cmdclass={"test": PyTest},
    version=about["__version__"],
    author=about["__maintainer__"],
    author_email=about["__email__"],
    url="http://github.com/GetStream/chat-py",
    description="Client for Stream Chat.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    zip_safe=False,
    install_requires=install_requires,
    extras_require={"test": tests_require, "ci": ci_require},
    tests_require=tests_require,
    include_package_data=True,
    python_requires=">=3.6",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Topic :: Software Development",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import uuid
import time
import pytest

from stream_chat.base.exceptions import StreamAPIException


@pytest.mark.incremental
class TestChannel(object):
    @pytest.mark.asyncio
    async def test_ban_user(self, event_loop, channel, random_user, server_user):
        await channel.ban_user(random_user["id"], user_id=server_user["id"])
        await channel.ban_user(
            random_user["id"],
            timeout=3600,
            reason="offensive language is not allowed here",
            user_id=server_user["id"],
        )
        await channel.unban_user(random_user["id"])

    @pytest.mark.asyncio
    async def test_create_without_id(self, event_loop, client, random_users):
        channel = client.channel(
            "messaging", data={"members": [u["id"] for u in random_users]}
        )
        assert channel.id is None

        await channel.create(random_users[0]["id"])
        assert channel.id is not None

    @pytest.mark.asyncio
    async def test_send_message_with_options(self, event_loop, channel, random_user):
        response = await channel.send_message(
            {"text": "hi"}, random_user["id"], skip_push=True
        )
        assert "message" in response
        assert response["message"]["text"] == "hi"

    @pytest.mark.asyncio
    async def test_send_event(self, event_loop, channel, random_user):
        response = await channel.send_event({"type": "typing.start"}, random_user["id"])
        assert "event" in response
        assert response["event"]["type"] == "typing.start"

    @pytest.mark.asyncio
    async def test_send_reaction(self, event_loop, channel, random_user):
        msg = await channel.send_message({"text": "hi"}, random_user["id"])
        response = await channel.send_reaction(
            msg["message"]["id"], {"type": "love"}, random_user["id"]
        )
        assert "message" in response
        assert len(response["message"]["latest_reactions"]) == 1
        assert response["message"]["latest_reactions"][0]["type"] == "love"

    @pytest.mark.asyncio
    async def test_delete_reaction(self, event_loop, channel, random_user):
        msg = await channel.send_message({"text": "hi"}, random_user["id"])
        await channel.send_reaction(
            msg["message"]["id"], {"type": "love"}, random_user["id"]
        )
        response = await channel.delete_reaction(
            msg["message"]["id"], "love", random_user["id"]
        )
        assert "message" in response
        assert len(response["message"]["latest_reactions"]) == 0

    @pytest.mark.asyncio
    async def test_update(self, event_loop, channel):
        response = await channel.update({"motd": "one apple a day..."})
        assert "channel" in response
        assert response["channel"]["motd"] == "one apple a day..."

    @pytest.mark.asyncio
    async def test_update_partial(self, event_loop, channel):
        response = await channel.update({"color": "blue", "age": 30})
        assert "channel" in response
        assert response["channel"]["color"] == "blue"
        assert response["channel"]["age"] == 30

        response = await channel.update_partial(
            to_set={"color": "red"}, to_unset=["age"]
        )
        assert "channel" in response
        assert response["channel"]["color"] == "red"
        assert "age" not in response["channel"]

    @pytest.mark.asyncio
    async def test_delete(self, event_loop, channel):
        response = await channel.delete()
        assert "channel" in response
        assert response["channel"].get("deleted_at") is not None

    @pytest.mark.asyncio
    async def test_truncate(self, event_loop, channel):
        response = await channel.truncate()
        assert "channel" in response

    @pytest.mark.asyncio
    async def test_add_members(self, event_loop, channel, random_user):
        response = await channel.remove_members([random_user["id"]])
        assert len(response["members"]) == 0

        response = await channel.add_members([random_user["id"]])
        assert len(response["members"]) == 1
        assert not response["members"][0].get("is_moderator", False)

    @pytest.mark.asyncio
    async def test_invite_members(self, event_loop, channel, random_user):
        response = await channel.remove_members([random_user["id"]])
        assert len(response["members"]) == 0

        response = await channel.invite_members([random_user["id"]])
        assert len(response["members"]) == 1
        assert response["members"][0].get("invited", True)

    @pytest.mark.asyncio
    async def test_add_moderators(self, event_loop, channel, random_user):
        response = await channel.add_moderators([random_user["id"]])
        assert response["members"][0]["is_moderator"]

        response = await channel.demote_moderators([random_user["id"]])
        assert not response["members"][0].get("is_moderator", False)

    @pytest.mark.asyncio
    async def test_mark_read(self, event_loop, channel, random_user):
        response = await channel.mark_read(random_user["id"])
        assert "event" in response
        assert response["event"]["type"] == "message.read"

    @pytest.mark.asyncio
    async def test_get_replies(self, event_loop, channel, random_user):
        msg = await channel.send_message({"text": "hi"}, random_user["id"])
        response = await channel.get_replies(msg["message"]["id"])
        assert "messages" in response
        assert len(response["messages"]) == 0

        for i in range(10):
            await channel.send_message(
                {"text": "hi", "index": i, "parent_id": msg["message"]["id"]},
                random_user["id"],
            )

        response = await channel.get_replies(msg["message"]["id"])
        assert "messages" in response
        assert len(response["messages"]) == 10

        response = await channel.get_replies(msg["message"]["id"], limit=3, offset=3)
        assert "messages" in response
        assert len(response["messages"]) == 3
        assert response["messages"][0]["index"] == 7

    @pytest.mark.asyncio
    async def test_get_reactions(self, event_loop, channel, random_user):
        msg = await channel.send_message({"text": "hi"}, random_user["id"])
        response = await channel.get_reactions(msg["message"]["id"])

        assert "reactions" in response
        assert len(response["reactions"]) == 0

        await channel.send_reaction(
            msg["message"]["id"], {"type": "love", "count": 42}, random_user["id"]
        )

        await channel.send_reaction(
            msg["message"]["id"], {"type": "clap"}, random_user["id"]
        )

        response = await channel.get_reactions(msg["message"]["id"])
        assert len(response["reactions"]) == 2

        response = await channel.get_reactions(msg["message"]["id"], offset=1)
        assert len(response["reactions"]) == 1

        assert response["reactions"][0]["count"] == 42

    @pytest.mark.asyncio
    async def test_send_and_delete_file(self, event_loop, channel, random_user):
        url = "helloworld.jpg"
        resp = await channel.send_file(url, "helloworld.jpg", random_user)
        assert "helloworld.jpg" in resp["file"]
        await channel.delete_file(resp["file"])

    @pytest.mark.asyncio
    async def test_send_and_delete_image(self, event_loop, channel, random_user):
        url = "helloworld.jpg"
        resp = await channel.send_image(
            url, "helloworld.jpg", random_user, content_type="image/jpeg"
        )
        assert "helloworld.jpg" in resp["file"]
        await channel.delete_image(resp["file"])

    @pytest.mark.asyncio
    async def test_channel_hide_show(self, event_loop, client, channel, random_users):
        # setup
        await channel.add_members([u["id"] for u in random_users])
        # verify
        response = await client.query_channels({"id": channel.id})
        assert len(response["channels"]) == 1
        response = await client.query_channels(
            {"id": channel.id}, user_id=random_users[0]["id"]
        )
        assert len(response["channels"]) == 1
        # hide
        await channel.hide(random_users[0]["id"])
        response = await client.query_channels(
            {"id": channel.id}, user_id=random_users[0]["id"]
        )
        assert len(response["channels"]) == 0
        # search hidden channels
        response = await client.query_channels(
            {"id": channel.id, "hidden": True}, user_id=random_users[0]["id"]
        )
        assert len(response["channels"]) == 1
        # unhide
        await channel.show(random_users[0]["id"])
        response = await client.query_channels(
            {"id": channel.id}, user_id=random_users[0]["id"]
        )
        assert len(response["channels"]) == 1
        # hide again
        await channel.hide(random_users[0]["id"])
        response = await client.query_channels(
            {"id": channel.id}, user_id=random_users[0]["id"]
        )
        assert len(response["channels"]) == 0
        # send message
        await channel.send_message({"text": "hi"}, random_users[1]["id"])
        # channel should be listed now
        response = await client.query_channels(
            {"id": channel.id}, user_id=random_users[0]["id"]
        )
        assert len(response["channels"]) == 1

    @pytest.mark.asyncio
    async def test_invites(self, event_loop, client, channel):
        members = ["john", "paul", "george", "pete", "ringo", "eric"]
        await client.update_users([{"id": m} for m in members])
        channel = client.channel(
            "team",
            "beatles-" + str(uuid.uuid4()),
            {"members": members, "invites": ["ringo", "eric"]},
        )
        await channel.create("john")
        # accept the invite when not a member
        with pytest.raises(StreamAPIException):
            await channel.accept_invite("brian")
        # accept the invite when a member
        accept = await channel.accept_invite("ringo")
        for m in accept["members"]:
            if m["user_id"] == "ringo":
                assert m["invited"] is True
                assert "invite_accepted_at" in m
        # can accept again, noop
        await channel.accept_invite("ringo")

        reject = await channel.reject_invite("eric")
        for m in reject["members"]:
            if m["user_id"] == "eric":
                assert m["invited"] is True
                assert "invite_rejected_at" in m
        # can reject again, noop
        await channel.reject_invite("eric")

    @pytest.mark.asyncio
    async def test_query_members(self, event_loop, client, channel):
        members = ["paul", "george", "john", "jessica", "john2"]
        await client.update_users([{"id": m, "name": m} for m in members])
        for member in members:
            await channel.add_members([member])

        response = await channel.query_members(
            filter_conditions={"name": {"$autocomplete": "j"}},
            sort=[{"field": "created_at", "direction": 1}],
            offset=1,
            limit=10,
        )

        assert len(response) == 2
        assert response[0]["user"]["id"] == "jessica"
        assert response[1]["user"]["id"] == "john2"

    @pytest.mark.asyncio
    async def test_mute_umute(self, event_loop, client, channel, random_users):
        user_id = random_users[0]["id"]
        response = await channel.mute(user_id, expiration=30000)
        assert "channel_mute" in response
        assert response["channel_mute"]["channel"]["cid"] == channel.cid
        assert response["channel_mute"]["user"]["id"] == user_id

        response = await client.query_channels(
            {"muted": True, "cid": channel.cid},
            user_id=user_id,
        )
        assert len(response["channels"]) == 1

        await channel.unmute(user_id)
        response = await client.query_channels(
            {"muted": True, "cid": channel.cid},
            user_id=user_id,
        )
        assert len(response["channels"]) == 0

    @pytest.mark.asyncio
    async def test_export_channel_status(self, client, channel):
        with pytest.raises(StreamAPIException, match=r".*Can't find task.*"):
            await client.get_export_channel_status(str(uuid.uuid4()))

        with pytest.raises(StreamAPIException, match=r".*Can't find channel.*"):
            await client.export_channel("messaging", str(uuid.uuid4()))

    @pytest.mark.asyncio
    async def test_export_channel(self, client, channel, random_users):
        await channel.send_message({"text": "Hey Joni"}, random_users[0]["id"])

        resp = await client.export_channel(channel.channel_type, channel.id)
        task_id = resp["task_id"]
        assert task_id != ""

        while True:
            resp = await client.get_export_channel_status(task_id)
            assert resp["status"] != ""
            assert resp["created_at"] != ""
            assert resp["updated_at"] != ""
            if resp["status"] == "completed":
                assert len(resp["result"]) != 0
                assert resp["result"]["url"] != ""
                assert len(resp["error"]) != 0
                break
            time.sleep(0.5)
